
import csv
import json
from justnotiondb.processors import processors
import requests
from typing import Self

class NotionClient:
    """
    A class representing a Notion API connection
    """
    url = 'https://api.notion.com/v1'
    def __init__(self: Self, secret: str) -> None:
        """
        Initializes a NotionAPI object with a secret, and sets up the
        headers for all requests. Also checks the API connection and raises
        an exception if the connection is invalid.

        Parameters
        ----------
        secret : str
            The secret for the Notion API.
        """
        self.__headers = {
            'Authorization': f'Bearer {secret}',
            'Notion-Version': '2022-06-28',
            'Content-Type': "application/json"
        }
        self.__error: None | str = None

    @property
    def headers(self: Self) -> dict:
        return self.__headers

    @property
    def error(self: Self) -> str | None:
        """
        The error message of the last failed request if any.
        
        Returns
        -------
        str | None
            The error message or None if the last request was successful.
        """
        return self.__error
    
    @error.setter
    def error(self: Self, error: str) -> None:
        self.__error = error

    def check(self: Self) -> bool:
        """
        Checks the connection to the Notion API by attempting to retrieve user information.
        
        Returns
        -------
        bool
            If the request is successful, returns True. 
            If the request fails, returns False.
        """
        try:
            response = requests.get(
                url=f'{self.url}/users',
                headers=self.headers
            )
            response.raise_for_status()
            return True
        except Exception as e:
            self.error = str(e)
            return False


class DB:
    """
    A class representing a Notion database
    """
    def __init__(self: Self, client: NotionClient, id: str) -> None:
        """
        Initializes a DB object.

        Parameters
        ----------
        client : NotionClient
            The NotionClient object to use for requests.

        id : str
            The ID of the database to query.
        """
        self.client = client
        self.id = id

    def fetch(self: Self, filter: dict) -> dict:
        """
        Queries the database and fetches the results as a JSON object.

        Parameters
        ----------
        filter : dict
            A filter to be applied to the query.
            You can find more information here: https://developers.notion.com/reference/post-database-query-filter

        Returns
        -------
        dict
            A JSON object containing the query results.
        """
        url = f"{self.client.url}/databases/{self.id}/query"
        response = requests.post(
            url, 
            headers=self.client.headers,
            data=json.dumps(filter)
        )
        response.raise_for_status()
        return response.json()

    def get(self: Self, filter: dict={}) -> list[dict]:
        """
        Queries the database and fetches the results as a list of dictionaries.

        Parameters
        ----------
        filter : dict, optional
            A filter to be applied to the query. Defaults to an empty filter.

        Returns
        -------
        list[dict]
            A list of dictionaries, each representing a page in the database.
            Each dictionary contains the page's properties, processed according to their type.
        """
        db = self.fetch(filter=filter)
        results = db['results']
        content = []
        for result in results:
            properties = result['properties']
            content.append({
                key: processors[value['type']](value)
                for key, value in properties.items()
            })
        return content

    @classmethod
    def write_csv(cls, content: list[dict], path: str, **kwargs) -> None:
        """
        Writes a list of dictionaries to a CSV file.

        Parameters
        ----------
        content : list[dict]
            A list of dictionaries, each representing a row in the CSV file
        path : str
            The path to the CSV file.
        **kwargs
            Additional keyword arguments to be passed to the CSV writer
        """
        with open(path, 'w', newline='', encoding='utf-8') as f:
            writer = csv.DictWriter(f, fieldnames=content[0].keys(), **kwargs)
            writer.writeheader()
            writer.writerows(content)

    @classmethod
    def write_json(cls, content: list[dict], path: str) -> None:
        """
        Writes a list of dictionaries to a JSON file.

        Parameters
        ----------
        content : list[dict]
            A list of dictionaries, each representing a row in the JSON file
        path : str
            The path to the JSON file
        """
        with open(path, 'w', encoding='utf-8') as f:
            json.dump(content, f, indent=4, ensure_ascii=False)




from typing import Callable

def process_date(property: dict) -> str | None:
    """
    Processes a date property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'date' field with
        'start' and optionally 'end' keys.

    Returns
    -------
    str
        A formatted string representing the date range if both
        start and end dates are present, otherwise just the start date.
    """
    value: dict = property['date']
    if value is None:
        return None
    start = value['start']
    end = value['end']
    result = f'{start} -> {end}' if end else start
    return result

def process_list(property: dict, field: str, subfield: str) -> str | None:
    value: list[dict] = property[field]
    results = [x[subfield] for x in value]
    if len(results):
        result = results if len(results) > 1 else results[0]
        if isinstance(result, list):
            return '||'.join(result)
        if isinstance(result, str):
            return result
        return None
    else:
        return None

def process_title(property: dict) -> str | None:
    """
    Processes a title property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'title' field, which is a list of text objects.

    Returns
    -------
    list[str]
        A list of plain text strings from the title field.
    """
    return process_list(property, 'title', 'plain_text')

def process_checkbox(property: dict) -> bool:
    """
    Processes a checkbox property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'checkbox' field with a boolean value.

    Returns
    -------
    bool
        The boolean value of the checkbox field.
    """
    return property['checkbox']

def process_rich_text(property: dict) -> str | None:
    """
    Processes a rich text property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'rich_text' field, which is a list of text objects.

    Returns
    -------
    list[str]
        A list of plain text strings from the rich text field.
    """
    return process_list(property, 'rich_text', 'plain_text')

def process_number(property: dict) -> int | None:
    """
    Processes a number property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'number' field with an integer value.

    Returns
    -------
    int
        The integer value of the number field.
    """
    return property['number']

def process_select(property: dict) -> str | None:
    """
    Processes a select property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'select' field with a dictionary value.

    Returns
    -------
    str
        The name of the selected option.
    """
    value: dict = property['select']
    if value is None:
        return None
    return value['name']

def process_multi_select(property: dict) -> str | None:
    """
    Processes a multi-select property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'multi_select' field, which is a list of option objects.

    Returns
    -------
    list[str]
        A list of names of the selected options.
    """
    return process_list(property, 'multi_select', 'name')

def process_status(property: dict) -> str | None:
    """
    Processes a status property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'status' field with a dictionary value.

    Returns
    -------
    str
        The name of the status.
    """
    value: dict = property['status']
    if value is None:
        return None
    return value['name']

def process_people(property: dict) -> str | None:
    """
    Processes a people property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'people' field, which is a list of people objects.

    Returns
    -------
    list[str]
        A list of names of the people.
    """
    return process_list(property, 'people', 'name')

def process_files(property: dict) -> str | None:
    """
    Processes a files property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'files' field, which is a list of file objects.

    Returns
    -------
    list[str]
        A list of URLs of the files.
    """
    return process_list(property, 'files', 'file')

def process_url(property: dict) -> str | None:
    """
    Processes a url property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'url' field with a string value.

    Returns
    -------
    str
        The string value of the url field.
    """
    return property['url']

def process_email(property: dict) -> str | None:
    """
    Processes an email property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'email' field with a string value.

    Returns
    -------
    str
        The string value of the email field.
    """
    return property['email']

def process_phone_number(property: dict) -> str | None:
    """
    Processes a phone number property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'phone_number' field with a string value.

    Returns
    -------
    str
        The string value of the phone number field.
    """
    return property['phone_number']

def process_formula(property: dict) -> str | None:
    """
    Processes a formula property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'formula' field with a dictionary value.

    Returns
    -------
    str
        The string value of the formula field.
    """
    value: dict = property['formula']
    result = value[value['type']]
    return result

def process_relation(property: dict) -> str | None:
    """
    Processes a relation property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'relation' field, which is a list of relation objects.

    Returns
    -------
    list[str]
        A list of IDs of the related pages.
    """
    return process_list(property, 'relation', 'id')

def process_rollup(property: dict) -> str | None:
    value: dict = property['rollup']
    value_type = value['type']
    if value_type == 'array':
        results = []
        for x in value[value_type]:
            results.append(str(processors[x['type']](x)))
        return '||'.join(results)
    else:
        return str(value[value_type])

def process_created_time(property: dict) -> str | None:
    """
    Processes a created time property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'created_time' field with a string value.

    Returns
    -------
    str
        The string value of the created time field.
    """
    return property['created_time']

def process_created_by(property: dict) -> str | None:
    """
    Processes a created by property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'created_by' field with a dictionary value.

    Returns
    -------
    str
        The string value of the created by field.
    """
    return property['created_by']['id']

def process_last_edited_time(property: dict) -> str | None:
    """
    Processes a last edited time property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'last_edited_time' field with a string value.

    Returns
    -------
    str
        The string value of the last edited time field.
    """
    return property['last_edited_time']

def process_last_edited_by(property: dict) -> str | None:
    """
    Processes a last edited by property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'last_edited_by' field with a dictionary value.

    Returns
    -------
    str
        The string value of the last edited by field.
    """
    return property['last_edited_by']['name']

def process_button(property: dict) -> None:
    """
    Processes a button property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'button' field with a dictionary value.

    Returns
    -------
    None
        The button property has no value to process, so this function simply returns None.
    """
    return None

def process_unique_id(property: dict) -> str | None:
    """
    Processes a unique id property from a Notion database entry.

    Parameters
    ----------
    property : dict
        A dictionary containing a 'unique_id' field with a string value.

    Returns
    -------
    str
        The string value of the unique id field.
    """
    return property['unique_id']['number']

processors: dict[str, Callable] = {
    'date': process_date,
    'title': process_title,
    'checkbox': process_checkbox,
    'rich_text': process_rich_text,
    'number': process_number,
    'select': process_select,
    'multi_select': process_multi_select,
    'status': process_status,
    'people': process_people,
    'files': process_files,
    'url': process_url,
    'email': process_email,
    'phone_number': process_phone_number,
    'formula': process_formula,
    'relation': process_relation,
    'rollup': process_rollup,
    'button': process_button,
    'created_time': process_created_time,
    'created_by': process_created_by,
    'last_edited_time': process_last_edited_time,
    'last_edited_by': process_last_edited_by,
    'unique_id': process_unique_id
}
